"""Concrete implementations of the Instruction Atom type."""
from abc import ABC, abstractmethod
from typing import Callable, Set, Sequence

from pyshgp.push.atoms import InstructionMeta
from pyshgp.push.config import PushConfig
from pyshgp.push.type_library import RESERVED_PSEUDO_STACKS
from pyshgp.push.state import PushState
from pyshgp.utils import Token


class Instruction(ABC):
    """A function in the Push language used to modify the PushState.

    The Instruction class is the abstract base class for specific implementations
    that are configured differently. For example, see SimpleInstruction verses
    TakesStateInstruction.

    Parameters
    ----------
    name : str,
        A unique name for the instruction.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    Attributes
    ----------
    name : str,
        A unique name for the instruction.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    """

    __slots__ = ["name", "code_block", "docstring"]

    def __init__(self, name: str, code_blocks: int, docstring="Write me!"):
        self.name = name
        self.code_blocks = code_blocks
        self.docstring = docstring

    @abstractmethod
    def evaluate(self, push_state: PushState, push_config: PushConfig = None) -> PushState:
        """Evaluate the instruction on the given PushState.

        Parameters
        ----------
        push_state: pyshgp.push.state.PushState
            The PushState to run the instruction on.
        push_config: pyshgp.push.interpreter.PushConfig
            The configuration of the Push language.

        Returns
        -------
        pyshgp.push.state.PushState
            Return the given state, possibly modified by the Instruction.

        """
        pass

    @abstractmethod
    def required_stacks(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction."""
        pass

    def meta(self) -> InstructionMeta:
        """Create an ``InstructionMeta`` from the instruction object."""
        return InstructionMeta(name=self.name, code_blocks=self.code_blocks)

    def __eq__(self, other):
        if type(self) == type(other):
            return self.name == other.name
        return False

    def __hash__(self):
        return self.name.__hash__()

    def __repr__(self):
        return "Instruction<{n}>".format(n=self.name)


class SimpleInstruction(Instruction):
    """A simple instruction implementation.

    A SimpleInstruction uses a standardized way of manipulating PushStates. In
    other words, it handles popping its own function arguments and pushing the
    function return values.

    The first step of evaluating a SimpleInstruction is to pop the arguments
    from the stacks corresponding the instruction's ``input_stacks`` list.
    If multiple occurrences of the same type are in ``input_stacks``, items are
    taken from progressively deeper in that stack. If the stacks of the
    PushState do not contain a sufficient number of items, the instruction does
    not modify the PushState.

    The popped arguments are then passed to the instruction's function to produce
    a tuple of outputs. It is crucial that the instruction's function produce a
    tuple of outputs, even if it only contains a single element. The elements of
    the tuple are then routed to the corresponding stacks specified in the
    instruction's ``output_stacks``.

    Parameters
    ----------
    name : str,
        A unique name for the instruction.
    f : Callable
        A function whose signature matches input_stacks and output_stacks.
    input_stacks : Sequence[str]
        A list of PushType names to use when popping arguments from the PushState.
    output_stacks : Sequence[str]
        A list of PushType names to use when pushing function results to the PushState.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    """

    __slots__ = ["name", "f", "input_stacks", "output_stacks", "code_blocks", "docstring"]

    def __init__(self,
                 name: str,
                 f: Callable,
                 input_stacks: Sequence[str],
                 output_stacks: Sequence[str],
                 code_blocks: int,
                 docstring="Write me!"):
        super().__init__(name, code_blocks, docstring)
        self.f = f
        self.input_stacks = input_stacks
        self.output_stacks = output_stacks

    def evaluate(self, push_state: PushState, push_config: PushConfig = None) -> PushState:
        """Evaluate the instruction on the given PushState. Return mutated State.

        A SimpleInstruction infers which values to pop and push from the stack
        based on its `input_stacks` and `output_stacks`.

        Parameters
        ----------
        push_state : PushState
            Push state to modify with the Instruction.
        push_config :  pyshgp.push.interpreter.PushConfig
            Configuration of the interpreter. Used to get various limits.

        Returns
        -------
        PushState
            Return the given state, possibly modified by the Instruction.

        """
        # Pull args, if present.
        args = push_state.observe_stacks(self.input_stacks)
        if Token.no_stack_item in args:
            return push_state

        # Compute result, return if revert or response too big.
        result = self.f(*args)
        if result is Token.revert:
            return push_state
        if not isinstance(result, (list, tuple)):
            raise ValueError("Instruction result must be a collection. {i} gave {t}.".format(
                i=self,
                t=type(result)
            ))

        # Remove arguments, push results.
        push_state.pop_from_stacks(self.input_stacks)
        push_state.push_to_stacks(result, self.output_stacks)
        return push_state

    def required_stacks(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction.

        Based on the the instructions input and output types.
        """
        return set(self.input_stacks + self.output_stacks) - RESERVED_PSEUDO_STACKS


class StateToStateInstruction(Instruction):
    """Instruction that takes entire PushState and returns entire PushState."""

    __slots__ = ["name", "f", "stacks_used", "code_blocks", "docstring"]

    def __init__(self,
                 name: str,
                 f: Callable,
                 stacks_used: Sequence[str],
                 code_blocks: int,
                 docstring="Write me!"):
        super().__init__(name, code_blocks, docstring)
        self.f = f
        self.stacks_used = set(stacks_used)

    def evaluate(self, push_state: PushState, push_config: PushConfig = None) -> PushState:
        """Evaluate the instruction on the given PushState. Return mutated State.

        A SimpleInstruction infers which values to pop and push from the stack
        based on its `input_stacks` and `output_stacks`.

        Parameters
        ----------
        push_state : PushState
            Push state to modify with the Instruction.
        push_config : PushConfig
            Configuration of the interpreter. Used to get various limits.

        Returns
        -------
        PushState
            Return the given state, possibly modified by the Instruction.

        """
        result = self.f(push_state)
        if result == Token.revert:
            return push_state
        else:
            return result

    def required_stacks(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction."""
        return self.stacks_used - RESERVED_PSEUDO_STACKS


class TakesStateInstruction(Instruction):
    """Instruction that takes entire PushState and returns particular values.

    The function of a TakesStateInstruction accepts an entire PushState as input
    and produces either a ``Token.revert`` or a tuple of outputs values. It is
    crucial that the instruction's function produce a tuple of outputs, even if
    it only conains a single element.

    The elements of the tuple are then routed to the corresponding stacks
    specified in the instruction's ``output_stacks``.

    Additional PushTypes utilized by the instruction are denoted in ``other_stacks``.

    Parameters
    ----------
    name : str,
        A unique name for the instruction.
    f : Callable
        A function that takes a PushState as input and produces values corresponding to ``output_stacks.``
    output_stacks : Sequence[str]
        A list of PushType names to use when pushing function results to the PushState.
    other_stacks : Sequence[str]
        A list of additional PushType names used by the Insturction's function.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    """

    __slots__ = ["name", "f", "output_stacks", "other_stacks", "code_blocks", "docstring"]

    def __init__(self,
                 name: str,
                 f: Callable,
                 output_stacks: Sequence[str],
                 other_stacks: Sequence[str],
                 code_blocks: int,
                 docstring="Write me!"):
        super().__init__(name, code_blocks, docstring)
        self.f = f
        self.output_stacks = output_stacks
        self.other_stacks = other_stacks

    def evaluate(self, push_state: PushState, push_config: PushConfig = None) -> PushState:
        """Evaluate the instruction on the given PushState. Return mutated State.

        A SimpleInstruction infers which values to pop and push from the stack
        based on its `input_stacks` and `output_stacks`.

        Parameters
        ----------
        push_state : PushState
            Push state to modify with the Instruction.
        push_config : PushConfig
            Configuration of the interpreter. Used to get various limits.

        Returns
        -------
        PushState
            Return the given state, possibly modified by the Instruction.

        """
        # Compute result. State should be modified in place during function.
        result = self.f(push_state)

        # Return if revert.
        if result is Token.revert:
            return push_state
        if not isinstance(result, (list, tuple)):
            raise ValueError("Instruction result must be a collection. {i} gave {t}.".format(
                i=self,
                t=type(result)
            ))

        # Push results.
        push_state.push_to_stacks(result, self.output_stacks)
        return push_state

    def required_stacks(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction."""
        return set(self.other_stacks + self.output_stacks) - RESERVED_PSEUDO_STACKS


class ProducesManyOfTypeInstruction(Instruction):
    """Instruction that produces arbitarily many values of a given PushType.

    ProducesManyOfTypeInstructions pop their arguments in the same was as
    SimpleInstructions. Items are popped from the stacks corresponding the
    types denoted in the ``input_stacks`` list. If multiple occurences of the
    same type are in ``input_stacks``, items are taken from progressively deeper
    in that stack. If the stacks of the PushState do not contain a sufficent
    number of items, the instruction does not modify the PushState.

    The popped arguments are then passed to the instruction's function to produce
    a tuple of outputs. It is crucial that the instruction's function produce a
    tuple of outputs, even if it only conains a single element. All elements of
    the tuple are pushed individually to the stack denoted in ``output_stack``.

    Parameters
    ----------
    name : str,
        A unique name for the instruction.
    f : Callable
        A function whose signature matches input_stacks and output_stacks.
    input_stacks : Sequence[str]
        A list of PushType names to use when popping arguments from the PushState.
    output_stack : str
        The name of a PushType to use when pushing function results to the PushState.
        All values returned by the function go to the stack for this type.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    """

    __slots__ = ["name", "f", "input_stacks", "output_stack", "code_blocks", "docstring"]

    def __init__(self,
                 name: str,
                 f: Callable,
                 input_stacks: Sequence[str],
                 output_stack: str,
                 code_blocks: int,
                 docstring="Write me!"):
        super().__init__(name, code_blocks, docstring)
        self.f = f
        self.input_stacks = input_stacks
        self.output_stack = output_stack

    def evaluate(self, push_state: PushState, push_config: PushConfig = None) -> PushState:
        """Evaluate the instruction on the given PushState. Return mutated State.

        A ProducesManyOfTypeInstruction infers which values to pop from the stack
        based on `input_stacks` and pushes each output to the same stack
        based on `output_stack`.

        Parameters
        ----------
        push_state : PushState
            Push state to modify with the Instruction.
        push_config : PushConfig
            Configuration of the interpreter. Used to get various limits.

        Returns
        -------
        PushState
            Return the given state, possibly modified by the Instruction.

        """
        # Pull args, if present.
        args = push_state.observe_stacks(self.input_stacks)
        if Token.no_stack_item in args:
            return push_state

        # Compute result, return if revert or response too big.
        result = self.f(*args)
        if result is Token.revert:
            return push_state
        if not isinstance(result, (list, tuple)):
            raise ValueError("Instruction result must be a collection. {i} gave {t}.".format(
                i=self,
                t=type(result)
            ))

        # Remove arguments, push results.
        push_state.pop_from_stacks(self.input_stacks)
        push_state.push_to_stacks(result, [self.output_stack] * len(result))
        return push_state

    def required_stacks(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction.

        Based on the the instructions input types and output type.
        """
        return set(self.input_stacks + [self.output_stack]) - RESERVED_PSEUDO_STACKS



"""Instructions common to all ``PushTypes`` and ``PushStacks``."""
from typing import Any, Tuple
from functools import partial

from pyshgp.push.types import PushType
from pyshgp.push.type_library import PushTypeLibrary
from pyshgp.push.instruction import (
    SimpleInstruction,
    TakesStateInstruction,
    StateToStateInstruction,
    ProducesManyOfTypeInstruction
)
from pyshgp.push.state import PushState
from pyshgp.push.atoms import Atom, Literal
from pyshgp.utils import Token


DUP_LIMIT = 500  # @todo move `DUP_LIMIT` to PushConfig


def _revert():
    return Token.revert


def _wrap_tuple(*args):
    return tuple(args)


def _noop(x=None):
    return tuple()


def _dup(x):
    return x, x


def _dup_times(times, item):
    x = int(min(times, DUP_LIMIT))
    return [item] * x


# Disabled due to performance issues.
# def _dup_top_n_factory(type_name: str) -> Callable:
#     def f(state: PushState) -> PushState:
#         if state["int"].is_empty() or state[type_name].is_empty():
#             return Token.revert
#         ndx = max(0, min(state["int"].pop(), len(state[type_name]) - 1))
#         for item in state[type_name][:ndx]:
#             state[type_name].push(item)
#         return state
#     return f


def _swap(a, b):
    return (a, b)


def _rot(a, b, c):
    return (b, a, c)


def _eq(a, b):
    return (a == b),


def _flush(state: PushState, type_name: str):
    state[type_name].flush()
    return state


def _stack_depth(state: PushState, type_name: str):
    return [len(state[type_name])]


def _yank(state: PushState, type_name: str) -> PushState:
    if state["int"].is_empty() or state[type_name].is_empty():
        return Token.revert
    if type_name == "int" and len(state[type_name]) < 2:
        return Token.revert
    raw_ndx = state["int"].pop()
    ndx = max(0, min(raw_ndx, len(state[type_name]) - 1))
    item = state[type_name].pop(ndx)
    state[type_name].push(item)
    return state


def _yank_dup(state: PushState, type_name: str):
    if state["int"].is_empty() or state[type_name].is_empty():
        return Token.revert
    if type_name == "int" and len(state[type_name]) < 2:
        return Token.revert
    raw_ndx = state["int"].pop()
    ndx = max(0, min(raw_ndx, len(state[type_name]) - 1))
    item = state[type_name].nth(ndx)
    state[type_name].push(item)
    return state


def _shove(state: PushState, type_name: str):
    if state["int"].is_empty() or state[type_name].is_empty():
        return Token.revert
    if type_name == "int" and len(state[type_name]) < 2:
        return Token.revert
    raw_ndx = state["int"].pop()
    ndx = max(0, min(raw_ndx, len(state[type_name]) - 1))
    item = state[type_name].pop()
    state[type_name].insert(ndx, item)
    return state


def _shove_dup(state: PushState, type_name: str):
    if state["int"].is_empty() or state[type_name].is_empty():
        return Token.revert
    if type_name == "int" and len(state[type_name]) < 2:
        return Token.revert
    raw_ndx = state["int"].pop()
    ndx = max(0, min(raw_ndx, len(state[type_name])))
    item = state[type_name].top()
    state[type_name].insert(ndx, item)
    return state


def _is_empty(state: PushState, type_name: str):
    return state[type_name].is_empty(),


def _make_code(x: Any, push_type: PushType) -> Tuple[Atom]:
    if isinstance(x, Atom):
        return x,
    else:
        return Literal(value=x, push_type=push_type),


def instructions(type_library: PushTypeLibrary):
    """Return all core numeric instructions."""
    i = []

    for push_type in type_library.keys():
        i.append(SimpleInstruction(
            "{t}_pop".format(t=push_type),
            _noop,
            input_stacks=[push_type],
            output_stacks=[],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Pops the top {t}.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_dup".format(t=push_type),
            _dup,
            input_stacks=[push_type],
            output_stacks=[push_type, push_type],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Duplicates the top {t}.".format(t=push_type)
        ))

        i.append(ProducesManyOfTypeInstruction(
            "{t}_dup_times".format(t=push_type),
            _dup_times,
            input_stacks=["int", push_type],
            output_stack=push_type,
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Duplicates the top {t} `n` times where `n` is from the int stack.".format(t=push_type)
        ))

        # Disabled due to performance issues.
        # i.append(StateToStateInstruction(
        #     "{t}_dup_top_n".format(t=push_type),
        #     _dup_top_n_factory(push_type),
        #     stacks_used=[push_type, "int"],
        #     code_blocks=0,
        #     docstring="Duplicates the top n items on the {t} stack.".format(t=push_type)
        # ))

        i.append(SimpleInstruction(
            "{t}_swap".format(t=push_type),
            _swap,
            input_stacks=[push_type, push_type],
            output_stacks=[push_type, push_type],
            code_blocks=(2 if push_type == "exec" else 0),
            docstring="Swaps the top two {t}s.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_rot".format(t=push_type),
            _rot,
            input_stacks=[push_type] * 3,
            output_stacks=[push_type] * 3,
            code_blocks=(3 if push_type == "exec" else 0),
            docstring="Rotates the top three {t}s.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_flush".format(t=push_type),
            partial(_flush, type_name=push_type),
            stacks_used=[push_type],
            code_blocks=0,
            docstring="Empties the {t} stack.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_eq".format(t=push_type),
            _eq,
            input_stacks=[push_type, push_type],
            output_stacks=["bool"],
            code_blocks=0,
            docstring="Pushes True if the top two {t} are equal. Otherwise pushes False.".format(t=push_type)
        ))

        i.append(TakesStateInstruction(
            "{t}_stack_depth".format(t=push_type),
            partial(_stack_depth, type_name=push_type),
            output_stacks=["int"],
            other_stacks=[push_type],
            code_blocks=0,
            docstring="Pushes the size of the {t} stack to the int stack.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_yank".format(t=push_type),
            partial(_yank, type_name=push_type),
            stacks_used=[push_type, "int"],
            code_blocks=0,
            docstring="Yanks a {t} from deep in the stack based on an index from the int stack and puts it on top.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_yank_dup".format(t=push_type),
            partial(_yank_dup, type_name=push_type),
            stacks_used=[push_type, "int"],
            code_blocks=0,
            docstring="Yanks a copy of a {t} deep in the stack based on an index from the int stack and puts it on top.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_shove".format(t=push_type),
            partial(_shove, type_name=push_type),
            stacks_used=[push_type, "int"],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Shoves the top {t} deep in the stack based on an index from the int stack.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_shove_dup".format(t=push_type),
            partial(_shove_dup, type_name=push_type),
            stacks_used=[push_type, "int"],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Shoves a copy of the top {t} deep in the stack based on an index from the int stack.".format(t=push_type)
        ))

        i.append(TakesStateInstruction(
            "{t}_is_empty".format(t=push_type),
            partial(_is_empty, type_name=push_type),
            output_stacks=["bool"],
            other_stacks=[push_type],
            code_blocks=0,
            docstring="Pushes True if the {t} stack is empty. Pushes False otherwise.".format(t=push_type)
        ))

    for push_type_name, push_type in type_library.items():
        if push_type_name == "code":
            continue
        i.append(SimpleInstruction(
            "code_from_{t}".format(t=push_type_name),
            partial(_make_code, push_type=push_type),
            input_stacks=[push_type_name],
            output_stacks=["code"],
            code_blocks=(1 if push_type_name == "exec" else 0),
            docstring="Moves the top {t} to the code stack.".format(t=push_type_name)
        ))

    return i



"""Definitions for all core I/O instructions, including input instructions."""
from pyshgp.push.instruction import SimpleInstruction
from pyshgp.push.type_library import PushTypeLibrary


# Printing instructions

def _wrap(x):
    return str(x),


def _wrap_and_newline(x):
    return str(x) + "\n",


def instructions(type_library: PushTypeLibrary):
    """Return all core printing instructions."""
    i = []

    for push_type in type_library.keys():
        i.append(SimpleInstruction(
            "print_{t}".format(t=push_type),
            _wrap,
            input_stacks=[push_type],
            output_stacks=["stdout"],
            code_blocks=0,
            docstring="Prints the top {t}.".format(t=push_type)
        )),
        i.append(SimpleInstruction(
            "println_{t}".format(t=push_type),
            _wrap_and_newline,
            input_stacks=[push_type],
            output_stacks=["stdout"],
            code_blocks=0,
            docstring="Prints the top {t}.".format(t=push_type)
        ))
    return i



"""Utility classes and functions used throughout pyshgp."""
from copy import deepcopy, copy
from enum import Enum
import inspect
import pickle
import os

import numpy as np


def instantiate_using(cls: type, args: dict):
    """Call the given function using only the relevant kwargs present in the args dict."""
    arg_names = inspect.getfullargspec(cls)[0][1:]
    kwargs = {}
    for arg_name in arg_names:
        if arg_name in args:
            kwargs[arg_name] = args[arg_name]
    return cls(**kwargs)


def list_rindex(lst, el):
    """Index of the last occurrence of an item in a list. Return None is item not found."""
    for i in reversed(range(len(lst))):
        if lst[i] == el:
            return i
    return None


class DiscreteProbDistrib:
    """Discrete Probability Distribution."""

    __slots__ = ["elements", "_total", "_raw_probabilities", "_normalized_probabilities"]

    def __init__(self):
        self.elements = []
        self._total = 0.0
        self._raw_probabilities = []
        self._normalize()

    def _normalize(self):
        self._normalized_probabilities = np.array(self._raw_probabilities) / self._total

    def add(self, el, p):
        """Add an element with a relative probability."""
        self.elements.append(el)
        self._total += float(p)
        self._raw_probabilities.append(p)
        self._normalize()
        return self

    def size(self):
        """Return the number of elements in the distribution."""
        return len(self.elements)

    def sample(self):
        """Return a sample from the distribution."""
        if self.size() == 1:
            return self.elements[0]
        return np.random.choice(self.elements, p=self._normalized_probabilities)

    def sample_n(self, n: int = 1, replace: bool = True):
        """Return n samples from the distribution."""
        if self.size() == 1 and replace:
            return [self.elements[0]] * n
        return np.random.choice(self.elements, n, replace, self._normalized_probabilities)


class Token(Enum):
    """Enum class of all Tokens."""

    no_stack_item = 1
    revert = 2
    whole_state = 3


class Saveable:
    """Allows a pickle-able class to be written and loaded from a file."""

    def save(self, path: str):
        """Save the CodeBlock to a binary file."""
        loc, filename = os.path.split(path)
        if loc != "":
            os.makedirs(loc, exist_ok=True)
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load(path: str):
        """Load a CodeBlock from a binary file.."""
        with open(path, "rb") as f:
            return pickle.load(f)


class Copyable:
    """Allows an object to be copied via a method."""

    def copy(self, deep: bool = False):
        """Copy the CodeBlock."""
        return deepcopy(self) if deep else copy(self)



"""The ``push`` sub-package contains an extensible Push interpreter and Push program abstraction.

When running Push programs (either during search or after) the modules in this package are used to perform the
stack-based execution of Push code and report output values. For more information on the Push programming language,
see the following links

- http://faculty.hampshire.edu/lspector/push3-description.html
- https://erp12.github.io/push-redux/

"""



"""The :mod:`estimator` module defines a ``PushEstimator`` class."""
from typing import Union, Tuple

from pyshgp.gp.individual import Individual

import pyshgp.gp.search as sr
import pyshgp.gp.selection as sl
import pyshgp.gp.variation as vr
from pyshgp.gp.evaluation import DatasetEvaluator
from pyshgp.gp.genome import GeneSpawner
from pyshgp.push.interpreter import PushInterpreter, DEFAULT_INTERPRETER
from pyshgp.push.config import PushConfig
from pyshgp.push.program import ProgramSignature
from pyshgp.tap import tap
from pyshgp.utils import list_rindex
from pyshgp.validation import check_is_fitted, check_X_y


class PushEstimator:
    """Simple estimator that synthesizes Push programs.

    Parameters
    ----------
    search : Union[SearchAlgorithm, str], optional
        The search algorithm, or its abbreviation, to use to when synthesizing
        Push programs.
    spawner : Union[GeneSpawner, str], optional
        The GeneSpawner to use when producing Genomes during initialization and
        variation. Default is all core intructions, no literals, and no ERC Generators.
    selector : Union[Selector, str], optional
        The selector, or name of selector, to use when selecting parents.
        The default is lexicase selection.
    variation_strategy : Union[VariationStrategy, dict, str]
        A VariationStrategy describing a collection of VariationOperators and how
        frequently to use them. If a dict is supplied, keys should be operator
        names and values should be the probability distirution. If a string is
        provided, the VariationOperators with that name will always be used.
    population_size : int, optional
        The number of individuals hold in the population each generation. Default
        is 300.
    max_generations : int, optional
        The number of generations to run the search algorithm. Default is 100.
    initial_genome_size : Tuple[int, int], optional
        The range of genome sizes to produce during initialization. Default is
        (20, 100)
    simplification_steps : int
        The number of simplification iterations to apply to the best Push program
        produced by the search algorithm.
    interpreter : PushInterpreter, optional
        The PushInterpreter to use when making predictions. Also holds the instruction
        set to use
    parallelism : Union[Int, bool], optional
        Set the number of processes to spawn for use when performing embarrassingly
        parallel tasks. If false, no processes will spawn and compuation will be
        serial. Default is true, which spawns one process per available cpu.
    verbose : int, optional
        Indicates if verbose printing should be used during searching.
        Default is 0. Options are 0, 1, or 2.
    **kwargs
        Arbitrary keyword arguments. Examples of supported arguments are
        `epsilon` (bool or float) when using Lexicase as the selector, and
        `tournament_size` (int) when using tournament selection.

    """

    def __init__(self,
                 spawner: GeneSpawner,
                 search: str = "GA",
                 selector: Union[sl.Selector, str] = "lexicase",
                 variation_strategy: Union[vr.VariationStrategy, dict, str] = "umad",
                 population_size: int = 300,
                 max_generations: int = 100,
                 initial_genome_size: Tuple[int, int] = (20, 100),
                 simplification_steps: int = 2000,
                 last_str_from_stdout: bool = False,
                 interpreter: PushInterpreter = "default",
                 parallelism: Union[int, bool] = False,
                 push_config: PushConfig = "default",
                 verbose: int = 0,
                 **kwargs):
        self._search_name = search
        self.spawner = spawner
        self.selector = selector
        self.variation_strategy = variation_strategy
        self.population_size = population_size
        self.max_generations = max_generations
        self.initial_genome_size = initial_genome_size
        self.simplification_steps = simplification_steps
        self.last_str_from_stdout = last_str_from_stdout
        self.parallelism = parallelism
        self.verbose = verbose
        self.ext = kwargs

        # Initialize attributes that will be set later.
        self.evaluator = None
        self.signature = None
        self.search = None
        self.solution = None

        if interpreter == "default":
            self.interpreter = DEFAULT_INTERPRETER
        else:
            self.interpreter = interpreter

        if push_config == "default":
            self.push_config = PushConfig()
        else:
            self.push_config = push_config

    def _build_search_algo(self):
        if isinstance(self.variation_strategy, dict):
            var_strat = vr.VariationStrategy()
            for op_name, prob in self.variation_strategy.items():
                var_op = vr.get_variation_operator(op_name)
                var_strat.add(var_op, prob)
            self.variation_strategy = var_strat

        search_config = sr.SearchConfiguration(
            signature=self.signature,
            spawner=self.spawner,
            evaluator=self.evaluator,
            selection=self.selector,
            variation=self.variation_strategy,
            population_size=self.population_size,
            max_generations=self.max_generations,
            initial_genome_size=self.initial_genome_size,
            simplification_steps=self.simplification_steps,
            parallelism=self.parallelism,
            push_config=self.push_config,
            verbose=self.verbose
        )
        self.search = sr.get_search_algo(self._search_name, config=search_config, **self.ext)

    @tap
    def fit(self, X, y):
        """Run the search algorithm to synthesize a push program.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The training input samples.
        y : list, array-like, or pandas dataframe.
            The target values (class labels in classification, real numbers in
            regression). Shape = [n_samples] or [n_samples, n_outputs]

        """
        X, y, arity, y_types = check_X_y(X, y)
        output_types = [self.interpreter.type_library.push_type_for_type(t).name for t in y_types]
        if self.last_str_from_stdout:
            ndx = list_rindex(output_types, "str")
            if ndx is not None:
                output_types[ndx] = "stdout"
        self.signature = ProgramSignature(arity=arity, output_stacks=output_types, push_config=self.push_config)
        self.evaluator = DatasetEvaluator(X, y, interpreter=self.interpreter)
        self._build_search_algo()
        self.solution = self.search.run()

    def predict(self, X):
        """Execute the synthesized push program on a dataset.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The set of cases to predict.

        Returns
        -------
        y_hat : pandas dataframe of shape = [n_samples, n_outputs]

        """
        check_is_fitted(self, "solution")
        return [
            self.interpreter.run(
                self.solution.program,
                inputs
            ) for inputs in X
        ]

    def score(self, X, y):
        """Run the search algorithm to synthesize a push program.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The training input samples.

        y : list, array-like, or pandas dataframe.
            The target values (class labels in classification, real numbers in
            regression). Shape = [n_samples] or [n_samples, n_outputs]

        """
        check_is_fitted(self, "solution")
        X, y, arity, y_types = check_X_y(X, y)
        self.evaluator = DatasetEvaluator(X, y, interpreter=self.interpreter)
        return self.evaluator.evaluate(self.solution.program)

    def save(self, filepath: str):
        """Load the found solution to a JSON file.

        Parameters
        ----------
        filepath
            Filepath to write the serialized search result to.

        """
        check_is_fitted(self, "solution")
        self.solution.save(filepath)

    def load(self, filepath: str):
        """Load a found solution from a JSON file.

        Parameters
        ----------
        filepath
            Filepath to read the serialized search result from.

        """
        self.solution = Individual.load(filepath)



"""The :mod:`genome` module defines the ``Genome`` type and provides genome translation, spawning and simplification.

A ``Genome`` is a persistent collection of gene Atoms (any `Atom` that isn't a ``CodeBlock``). It can be translated
into a ``CodeBlock``.

The ``GeneSpawner`` is a factory capable of generating random genes and random genomes. It is used for initializing a
population as well producing new genes used by variation operators (ie. mutation).

The genome simplification process is useful for removing superfluous genes from a genome without negatively impacting
the behavior of the program produced by the genome. This process has many benefits including: improving generalization,
shrinking the size of the serialized solution, and in some cases making the program easier to explain.

"""
from __future__ import annotations

from enum import Enum
from typing import Sequence, Union, Any, Callable, Tuple

import numpy as np
from pyrsistent import PRecord, field, CheckedPVector, l

from pyshgp.gp.evaluation import Evaluator
from pyshgp.push.instruction_set import InstructionSet
from pyshgp.push.program import ProgramSignature, Program
from pyshgp.push.atoms import Atom, CodeBlock, Closer, Literal, InstructionMeta, Input
from pyshgp.push.type_library import infer_literal
from pyshgp.tap import tap
from pyshgp.utils import DiscreteProbDistrib


class Opener(PRecord):
    """Marks the start of one or more CodeBlock."""

    count = field(type=int, mandatory=True)

    def dec(self) -> Opener:
        """Create an ``Opener`` with ``count`` decremented."""
        return Opener(count=self.count - 1)


def _has_opener(seq: Sequence) -> bool:
    for el in seq:
        if isinstance(el, Opener):
            return True
    return False


class Genome(CheckedPVector):
    """A linear sequence of genes (aka any atom that isn't a ``CodeBlock``).

    PyshGP uses the Plushy genome representation.

    See: http://gpbib.cs.ucl.ac.uk/gp-html/Spector_2019_GPTP.html

    """

    __type__ = Atom
    __invariant__ = lambda a: (not isinstance(a, CodeBlock), 'CodeBlock')


def genome_to_code(genome: Genome) -> CodeBlock:
    """Translate into nested CodeBlocks.

    These CodeBlocks can be considered the Push program representation of
    the Genome which can be executed by a PushInterpreter and evaluated
    by an Evaluator.

    """
    plushy_buffer = l()
    for atom in genome[::-1]:
        if isinstance(atom, InstructionMeta) and atom.code_blocks > 0:
            plushy_buffer = plushy_buffer.cons(Opener(count=atom.code_blocks))
        plushy_buffer = plushy_buffer.cons(atom)

    push_buffer = []
    while True:
        # If done with plush but unclosed opens, recur with one more close.
        if len(plushy_buffer) == 0 and _has_opener(push_buffer):
            plushy_buffer = plushy_buffer.cons(Closer())
        # If done with plush and all opens closed, return push.
        elif len(plushy_buffer) == 0:
            return CodeBlock(push_buffer)
        else:
            atom = plushy_buffer.first
            plushy_buffer = plushy_buffer.rest
            # If next instruction is a close, and there is an open.
            if isinstance(atom, Closer) and _has_opener(push_buffer):
                ndx, opener = [(ndx, el) for ndx, el in enumerate(push_buffer) if isinstance(el, Opener)][-1]
                post_open = push_buffer[ndx + 1:]
                pre_open = push_buffer[:ndx]
                push_buffer = pre_open + [CodeBlock(post_open)]
                if opener.count > 1:
                    opener = opener.dec()
                    push_buffer.append(opener)
            # If next instruction is a close, and there is no open, ignore it.
            # If next instruction is not a close.
            elif not isinstance(atom, Closer):
                push_buffer.append(atom)


class GeneTypes(Enum):
    """An ``Enum`` denoting the different types of genes that can appear in a Genome."""

    INPUT = 1
    INSTRUCTION = 2
    CLOSE = 3
    LITERAL = 4
    ERC = 5


class GeneSpawner:
    """A factory of random Genes (Atoms) and Genomes.

    When  spawning a random gene, the result can be one of three types of Atoms.
    An Instruction, a Closer, or a Literal. If the Atom is a Literal, it may
    be one of the supplied Literals, or it may be the result of running one of
    the Ephemeral Random Constant generators.

    Reference for ERCs:
    "A field guide to genetic programming", Section 3.1
    Riccardo Poli and William B. Langdon and Nicholas Freitag McPhee,
    http://www.gp-field-guide.org.uk/

    Attributes
    ----------
    n_input : int
        Number of input instructions that could appear the genomes.
    instruction_set : pyshgp.push.instruction_set.InstructionSet
        InstructionSet containing instructions to use when spawning genes and
        genomes.
    literals : Sequence[pyshgp.push.instruction_set.atoms.Literal]
        A list of Literal objects to pull from when spawning genes and genomes.
    erc_generator : Sequence[Callable]
        A list of functions (aka Ephemeral Random Constant generators). When one
        of these functions is called, the output is placed in a Literal and
        returned as the spawned gene.
    distribution : pyshgp.utils.DiscreteProbDistrib
        A probability distribution describing how frequently to produce
        Instructions, Closers, Literals, and ERCs.

    """

    def __init__(self,
                 n_inputs: int,
                 instruction_set: Union[InstructionSet, str],
                 literals: Sequence[Any],
                 erc_generators: Sequence[Callable],
                 distribution: DiscreteProbDistrib = "proportional"):
        self.n_inputs = n_inputs
        self.erc_generators = erc_generators

        self.instruction_set = instruction_set
        if self.instruction_set == "core":
            self.instruction_set = InstructionSet(register_core=True)
        self.type_library = self.instruction_set.type_library
        self.literals = [lit if isinstance(lit, Literal) else infer_literal(lit, self.type_library) for lit in literals]

        if distribution == "proportional":
            self.distribution = (
                DiscreteProbDistrib()
                .add(GeneTypes.INPUT, self.n_inputs)
                .add(GeneTypes.INSTRUCTION, len(self.instruction_set))
                .add(GeneTypes.CLOSE, sum([i.code_blocks for i in self.instruction_set.values()]))
                .add(GeneTypes.LITERAL, len(literals))
                .add(GeneTypes.ERC, len(erc_generators))
            )
        else:
            self.distribution = distribution

    def random_input(self) -> Input:
        """Return a random ``Input``.

        Returns
        -------
        pyshgp.push.atoms.Input

        """
        return Input(input_index=np.random.randint(self.n_inputs))

    def random_instruction(self) -> InstructionMeta:
        """Return a random Instruction from the InstructionSet.

        Returns
        -------
        pyshgp.push.atoms.InstructionMeta
            A randomly selected Literal.

        """
        i = np.random.choice(list(self.instruction_set.values()))
        return InstructionMeta(name=i.name, code_blocks=i.code_blocks)

    def random_literal(self) -> Literal:
        """Return a random Literal from the set of Literals.

        Returns
        -------
        pyshgp.push.atoms.Literal
            A randomly selected Literal.

        """
        lit = np.random.choice(self.literals)
        if not isinstance(lit, Literal):
            lit = infer_literal(lit, self.type_library)
        return lit

    def random_erc(self) -> Literal:
        """Materialize a random ERC generator into a Literal and return it.

        Returns
        -------
        pyshgp.push.atoms.Literal
            A Literal whose value comes from running a ERC generator function.

        """
        erc_value = np.random.choice(self.erc_generators)()
        if not isinstance(erc_value, Literal):
            erc_value = infer_literal(erc_value, self.type_library)
        return erc_value

    def random_gene(self) -> Atom:
        """Return a random Atom based on the GenomeSpawner's distribution.

        Returns
        -------
        pyshgp.push.atoms.Atom
            An random Atom. Either an Instruction, Closer, or Literal.

        """
        atom_type = self.distribution.sample()
        if atom_type is GeneTypes.INPUT:
            return self.random_input()
        elif atom_type is GeneTypes.INSTRUCTION:
            return self.random_instruction()
        elif atom_type is GeneTypes.CLOSE:
            return Closer()
        elif atom_type is GeneTypes.LITERAL:
            return self.random_literal()
        elif atom_type is GeneTypes.ERC:
            return self.random_erc()
        else:
            raise ValueError("GenomeSpawner distribution bad atom type {t}".format(t=str(atom_type)))

    def spawn_genome(self, size: Union[int, Sequence[int]]) -> Genome:
        """Return a random Genome based on the GenomeSpawner's distribution.

        The genome will contain the specified number of Atoms if size is an
        integer. If size is a pair of integers, the genome will be of a random
        size in the range of the two integers.

        Parameters
        ----------
        size
            The resulting genome will contain this many Atoms if size is an
            integer. If size is a pair of integers, the genome will be of a random
            size in the range of the two integers.

        Returns
        -------
        pyshgp.gp.genome.Genome
            A Genome with random contents of a given size.

        """
        if isinstance(size, Sequence):
            size = np.random.randint(size[0], size[1]) + 1
        return Genome([self.random_gene() for _ in range(size)])


class GenomeSimplifier:
    """Simplifies a genome while preserving, or improving, its error.

    Genomes, and Push programs, can contain superfluous Push code. This extra
    code often has no effect on the program behavior, but occasionally it can
    introduce subtle errors or behaviors that is not covered by the training
    cases. Removing the superfluous code makes genomes (and thus programs)
    smaller and easier to understand. More importantly, simplification can
    improve the generalization of the given genome/program.

    The process of genome simplification is iterative and closely resembles
    simple hill climbing. For each iteration, the simplifier will randomly
    select a small number of random genes to remove. The Genome is re-evaluated
    and if its error gets worse, the change is reverted. After repeating this
    for some number of steps, the resulting genome will be the same size or
    smaller while containing the same (or better) error value.

    Reference:
    "Improving generalization of evolved programs through automatic simplification"
    Thomas Helmuth, Nicholas Freitag McPhee, Edward Pantridge, and Lee Spector. 2017.
    In Proceedings of the Genetic and Evolutionary Computation Conference (GECCO '17).
    ACM, New York, NY, USA, 937-944. DOI: https://doi.org/10.1145/3071178.3071330

    See: https://dl.acm.org/citation.cfm?id=3071178.3071330

    """

    # @TODO: Add noop swaps to simplification.

    def __init__(self,
                 evaluator: Evaluator,
                 program_signature: ProgramSignature):
        self.evaluator = evaluator
        self.program_signature = program_signature

    def _remove_rand_genes(self, genome: Genome) -> Genome:
        # @todo DRY with deletion variation operator.
        gn = genome
        n_genes_to_remove = min(np.random.randint(1, 4), len(genome) - 1)
        ndx_of_genes_to_remove = np.random.choice(np.arange(len(gn)), n_genes_to_remove, replace=False)
        ndx_of_genes_to_remove[::-1].sort()
        for ndx in ndx_of_genes_to_remove:
            gn = gn.delete(ndx)
        return gn

    def _errors_of_genome(self, genome: Genome) -> np.ndarray:
        cb = genome_to_code(genome)
        program = Program(code=cb, signature=self.program_signature)
        return self.evaluator.evaluate(program)

    @tap
    def _step(self, genome: Genome, errors_to_beat: np.ndarray) -> Tuple[Genome, np.ndarray]:
        new_gn = self._remove_rand_genes(genome)
        new_errs = self._errors_of_genome(new_gn)
        if np.sum(new_errs) <= np.sum(errors_to_beat):
            return new_gn, new_errs
        return genome, errors_to_beat

    @tap
    def simplify(self,
                 genome: Genome,
                 original_errors: np.ndarray,
                 steps: int = 2000) -> Tuple[Genome, np.ndarray]:
        """Simplify the given genome while maintaining error.

        Parameters
        ----------
        genome
            The Genome to simplify.
        original_errors
            Error vector of the genome to simplify.
        steps
            Number of simplification iterations to perform. Default is 2000.

        Returns
        -------
        pyshgp.gp.genome.Genome
            The shorter Genome that expresses the same computation.

        """
        gn = genome
        errs = original_errors
        for step in range(steps):
            gn, errs = self._step(gn, errs)
            if len(gn) == 1:
                break
        return gn, errs



"""The :mod:`population` module defines an evolutionary population of Individuals."""
from collections.abc import Sequence
from bisect import insort_left
import numpy as np
import pickle
from multiprocessing import Pool
from functools import partial

from pyshgp.gp.individual import Individual
from pyshgp.gp.evaluation import Evaluator
from pyshgp.tap import tap


def _eval_indiv(indiv: Individual, evalr: Evaluator, ):
    indiv.error_vector = evalr.evaluate(indiv.program)
    return indiv


class Population(Sequence):
    """A sequence of Individuals kept in sorted order, with respect to their total errors."""

    __slots__ = ["unevaluated", "evaluated"]

    def __init__(self, individuals: list = None):
        self.unevaluated = []
        self.evaluated = []

        if individuals is not None:
            for el in individuals:
                self.add(el)

    def __len__(self):
        return len(self.evaluated) + len(self.unevaluated)

    def __getitem__(self, key: int) -> Individual:
        if key < len(self.evaluated):
            return self.evaluated[key]
        return self.unevaluated[key - len(self.evaluated)]

    def add(self, individual: Individual):
        """Add an Individaul to the population."""
        if individual.total_error is None:
            self.unevaluated.append(individual)
        else:
            insort_left(self.evaluated, individual)
        return self

    def best(self):
        """Return the best n individual in the population."""
        return self.evaluated[0]

    def best_n(self, n: int):
        """Return the best n individuals in the population."""
        return self.evaluated[:n]

    @tap
    def p_evaluate(self, evaluator_proxy, pool: Pool):
        """Evaluate all unevaluated individuals in the population in parallel."""
        func = partial(_eval_indiv, evalr=evaluator_proxy)
        for individual in pool.imap_unordered(func, self.unevaluated):
            insort_left(self.evaluated, individual)
        self.unevaluated = []

    @tap
    def evaluate(self, evaluator: Evaluator):
        """Evaluate all unevaluated individuals in the population."""
        for individual in self.unevaluated:
            individual = _eval_indiv(individual, evaluator)
            insort_left(self.evaluated, individual)
        self.unevaluated = []

    def all_error_vectors(self):
        """2D array containing all Individuals' error vectors."""
        return np.array([i.error_vector for i in self.evaluated])

    def all_total_errors(self):
        """1D array containing all Individuals' total errors."""
        return np.array([i.total_error for i in self.evaluated])

    def median_error(self):
        """Median total error in the population."""
        return np.median(self.all_total_errors())

    def error_diversity(self):
        """Proportion of unique error vectors."""
        return len(np.unique(self.all_error_vectors(), axis=0)) / float(len(self))

    def genome_diversity(self):
        """Proportion of unique genomes."""
        unq = set([pickle.dumps(i.genome) for i in self])
        return len(unq) / float(len(self))

    def program_diversity(self):
        """Proportion of unique programs."""
        unq = set([pickle.dumps(i.program.code) for i in self])
        return len(unq) / float(len(self))

    def mean_genome_length(self):
        """Average genome length across all individuals."""
        tot_gn_len = sum([len(i.genome) for i in self])
        return tot_gn_len / len(self)



"""The ``Tap`` and ``TapManager`` abstractions allow for users to inject side effects into PyshGP runs.

These side effects could be print statements for monitoring, writing to log files, or sending messages to
external systems when certain events occur. A ``Tap`` can inject a side effect before a function call, after a
function call, or both. Any function or method defined in PyshGP or your application code can be tapped with the
``@tap`` decorator defined in the module.

The taps provided in PyshGP are sufficient for most use cases, but user can define and register their
own ``Tap`` objects via inheritance and method overrides.

"""
import csv
import inspect
import json
import os
from datetime import datetime

import numpy as np
from abc import ABC
from functools import wraps
from typing import Dict, Sequence, Optional, MutableMapping


class Tap(ABC):
    """A debugging/logging abstraction to a function and collects its arguments and returned.

    The methods of a ``Tap`` should be side effects. It is not recommended to make state changes from with a ``Tap`` if
    that state will change code behavior.
    """

    def pre(self, id: str, args, kwargs):
        """Perform a particular side-effect directly before the associated function/method is called.

        Parameters
        ----------
        id : str
            The ID of the tap, generated based off of the name qualified name of the function.
        args : list
            The positional of the function call. For methods, ``args[0]`` is the class instance.
        kwargs : dict
            The keyword args of the function call.

        """
        pass

    def post(self, id: str, args, kwargs, returned):
        """Perform a particular side-effect directly before the associated function/method is called.

        Parameters
        ----------
        id : str
            The ID of the tap, generated based off of the name qualified name of the function.
        args : list
            The positional of the function call. For methods, ``args[0]`` is the class instance.
        kwargs : dict
            The keyword args of the function call.
        returned : Any
            The returned value of the function call.

        """
        pass

    def do(self, id: str, *args, **kwargs):
        """Perform a particular side-effect when called.

        Parameters
        ----------
        id : str
            The ID of the tap.

        """
        pass


class LogFileTap(Tap):
    """A ``Tap`` for writing log files.

    Parameters
    ----------
    root : str
        The root directory for putting log files (and directories).

    """

    # @todo Make log timestamps configurable.

    def __init__(self, root: str):
        self.root = root

    def dir(self, id: str) -> str:
        """Generate the directory for log files associated with the given function ID."""
        return os.path.join(self.root, id.replace(".", os.path.sep))

    def path(self, id: str, filename: str) -> str:
        """Generate the path for the log files with the given name under the directory for the given function ID."""
        return os.path.join(self.dir(id), filename)

    def log(self, id: str, filename: str, line: str):
        """Write a line to a log file. Includes a timestamp for each log entry.

        Parameters
        ----------
        id : str
            The ID of the function being tapped. This is used to generate a directory for the log file.
        filename : str
            The name of the log file.
        line : str
            The line of text to write to the log file.

        """
        path = self.path(id, filename)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(self.path(id, filename), "a") as f:
            f.write(str(datetime.now()))
            f.write("\t")
            f.write(line)
            f.write("\n")


class CsvTap(LogFileTap):
    """A ``Tap`` for writing CSV files.

    Parameters
    ----------
    root : str
        The root directory for putting log files (and directories).
    column_names : Sequence[str]
        The column names to put in the header of the CSV file.

    """

    def __init__(self, root: str, column_names: Sequence[str]):
        super().__init__(root)
        self.column_names = column_names

    def log(self, id: str, filename: str, row: Dict):
        """Write a to a CSV file.

        Parameters
        ----------
        id : str
            The ID of the function being tapped. This is used to generate a directory for the log file.
        filename : str
            The name of the CSV file.
        row : dict
            A dictionary that will be written as a row in the CSV. The keys should be in the
            CsvTap's list of ``column_names``.

        """
        path = self.path(id, filename)
        is_new = not os.path.exists(path)
        if is_new:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "a") as f:
            writer = csv.DictWriter(f, fieldnames=self.column_names)
            if is_new:
                writer.writeheader()
            writer.writerow(row)


class JsonLinesTap(LogFileTap):
    """A ``Tap`` for writing JSON lines files.

    Parameters
    ----------
    root : str
        The root directory for putting log files (and directories).

    """

    def __init__(self, root: str):
        super().__init__(root)

    def log(self, id: str, filename: str, row: Dict):
        """Write a JSON object to a line of a JSON lines file.

        Parameters
        ----------
        id : str
            The ID of the function being tapped. This is used to generate a directory for the log file.
        filename : str
            The name of the JSON file.
        row : dict
            A dictionary that will be written as a row in the file.

        """
        path = self.path(id, filename)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "a") as f:
            f.writelines([json.dumps(row)])


class StdOutRun(Tap):
    """A ``Tap`` that prints information from a ``SearchAlgorithm`` run to stdout.

    Parameters
    ----------
    pre_print_config : bool, optional
        If True, print the PushConfig before each run. Default False.
    pre_print_atoms : bool, optional
        If True, print all the atoms used to generate programs before each run. Default False.
    post_print_best : bool, optional
        If True, print a summary about the best program found at the end of each run. Default False.

    """

    def __init__(self, *,
                 pre_print_config: bool = False,
                 pre_print_atoms: bool = False,
                 post_print_best: bool = False):
        self.pre_print_config = pre_print_config
        self.pre_print_atoms = pre_print_atoms
        self.post_print_best = post_print_best

    def pre(self, id: str, args, kwargs, obj=None):
        """Print run config and/or all atoms to stdout."""
        search = args[0]
        if self.pre_print_config:
            print("Search Configuration:")
            attrs = ["signature", "evaluator", "spawner", "population_size", "max_generations", "error_threshold",
                     "initial_genome_size", "simplification_steps", "parallel_context", "selection", "variation"]
            for attr in attrs:
                print(attr + ":", getattr(search.config, attr))
        if self.pre_print_atoms:
            print("Instructions:")
            print(", ".join([i for i in search.config.spawner.instruction_set.keys()]))
            print("Literals:")
            print(search.config.spawner.literals)
            print("ERC Generators:")
            print(search.config.spawner.erc_generators)

    def post(self, id: str, args, kwargs, returned, obj=None):
        """Print a summary of the run result to stdout."""
        search = args[0]
        if search.is_solved():
            print("Solution found.")
        else:
            print("No solution found.")

        if self.post_print_best:
            print("Best individual found:")
            print("Genome:", search.best_seen.genome)
            print("Program:", search.best_seen.program)
            print("Error vector:", search.best_seen.error_vector)
            print("Total error:", search.best_seen.total_error)


class StdOutSearchStepTap(Tap):
    """A ``Tap`` that prints stats from a step (aka generation) of a ``SearchAlgorithm`` run."""

    def __init__(self, every_n_steps: int):
        self.every_n_steps = every_n_steps

    def pre(self, id: str, args, kwargs, obj=None):
        """Print population stats before the next step of the run."""
        search = args[0]
        generation = search.generation
        if generation % self.every_n_steps == 0:
            print(" | ".join([
                str(datetime.now()),
                "GENERATION: {g}".format(
                    g=generation
                ),
                "ERRORS: best={b}, median={m}, diversity={e_d}".format(
                    b=search.population.best().total_error,
                    m=search.population.median_error(),
                    e_d=search.population.error_diversity()
                ),
                "INDIVIDUALS: n={ps}, avg_genome_length={gn_len}".format(
                    ps=len(search.population),
                    gn_len=search.population.mean_genome_length()
                ),
            ]))


class StdOutSimplification(Tap):
    """A ``Tap`` that prints a summary of Genome simplification."""

    def pre(self, id: str, args, kwargs, obj=None):
        """Print a notification that genome simplification is starting."""
        genome = args[1]
        print("Simplifying genome of length {ln} with total error {er}.".format(
            ln=len(genome),
            er=np.sum(args[2])
        ))

    def post(self, id: str, args, kwargs, returned, obj=None):
        """Print a summary of the result of genome simplification."""
        gn, errs = returned
        print("Simplified genome to length {ln} and total error {er}.".format(
            ln=len(gn),
            er=np.sum(errs.sum())
        ))


class StdOutSimplificationStep(Tap):
    """A ``Tap`` that prints a notification of a successful Genome simplification step."""

    def post(self, id: str, args, kwargs, returned, obj=None):
        """Print a notification of a successful step of Genome simplification."""
        orig_gn = args[1]
        new_gn = returned[0]
        orig_len = len(orig_gn)
        new_len = len(new_gn)
        if new_len < orig_len:
            print("Simplified genome by {diff} to length {ln}.".format(
                diff=orig_len - new_len,
                ln=new_len
            ))


class TapManager:
    """Stores a mapping of function ID to ``Tap`` object than can be used to inject side effects around functions.

    The ``TapManger`` class is treated as a singleton and should not be instanced. Its methods are static and the state
    they manage is shared between all usages of ``Taps``.

    Function IDs are a fully qualified identifier for a function. They are generated as the concatenation
    of the module name the function is defined in (ie. ``my_package.my_module``) and the qualified name of the function
    definition (ie. ``MyClass.my_method``). The final function ID would be ``my_package.my_module.MyClass.my_method``.

    """

    _taps: MutableMapping[str, Tap] = {}

    # @todo multiple Taps per ID? CompositeTap type?

    @staticmethod
    def register(id: str, tap: Tap):
        """Register a ``Tap`` to be performed when the function with the associated ID is called."""
        TapManager._taps[id] = tap

    @staticmethod
    def unregister(id: str):
        """Unregister the ``Tap`` associated with given ID."""
        if id in TapManager._taps:
            del TapManager._taps[id]

    @staticmethod
    def get(id: str) -> Optional[Tap]:
        """Return the ``Tap`` associated with given ID or ``None`` if no tap is registered."""
        if id in TapManager._taps:
            return TapManager._taps[id]
        return None

    @staticmethod
    def do(id: str, *args, **kwargs):
        """Perform the ``do`` method of the tap registered under the given ID.

        This can be called from any location.
        """
        if id in TapManager._taps:
            TapManager._taps[id].do(id, *args, **kwargs)
        else:
            raise KeyError("No Tap registered in TapManager with id " + id)


def tap(fn):
    """Decorate a function/method to call any associated taps that have been registered in the ``TapManager``.

    Functional behavior is not changed.

    """
    fn_id = inspect.getmodule(fn).__name__ + "." + fn.__qualname__

    @wraps(fn)
    def tapped(*args, **kwargs):
        tap = TapManager.get(fn_id)
        if tap is not None:
            tap.pre(fn_id, args, kwargs)
        result = fn(*args, **kwargs)
        if tap is not None:
            tap.post(fn_id, args, kwargs, result)
        return result

    return tapped


def set_verbosity(level: int):
    """Register some ``Tap`` objects in the ``TapManger`` that print to stdout during an evolutionary run.

    Verbosity level 0 prints nothing to stdout.

    Verbosity level 1 prints minimal run configuration as well as population statistics every 5 generations.

    Verbosity level 2+ prints full run configuration, population statics every generation, and traces the genome
    simplification process.

    """
    if level > 0:
        TapManager.register("pyshgp.gp.genome.GenomeSimplifier.simplify", StdOutSimplification())

    if level == 1:
        TapManager.register("pyshgp.gp.search.SearchAlgorithm.step", StdOutSearchStepTap(every_n_steps=5))
        TapManager.register("pyshgp.gp.search.SearchAlgorithm.run", StdOutRun(pre_print_config=True,
                                                                              pre_print_atoms=False,
                                                                              post_print_best=False))
    if level > 1:
        TapManager.register("pyshgp.gp.search.SearchAlgorithm.step", StdOutSearchStepTap(every_n_steps=1))
        TapManager.register("pyshgp.gp.search.SearchAlgorithm.run", StdOutRun(pre_print_config=True,
                                                                              pre_print_atoms=True,
                                                                              post_print_best=True))
        TapManager.register("pyshgp.gp.genome.GenomeSimplifier._step", StdOutSimplificationStep())



"""The :mod:`program` module defines an abstraction to represent Push programs and their the specification.

Program objects encapsulate everything required to execute the program on a Push interpreter with the exception
of instruction definitions. Programs are serializable, and thus can be saved and reused. There is the possibility
(but not a guarantee) that Push programs can be executed across different versions of ``pyshgp`.

"""
from pyrsistent import PRecord, field

from pyshgp.push.config import PushConfig
from pyshgp.push.atoms import CodeBlock
from pyshgp.utils import Saveable


class ProgramSignature(PRecord):
    """A specification of a Push program.

    Attributes
    ----------
    arity : int
        The number of inputs that the program will take.
    output_stacks : List[str]
        The names of the stack(s) which the output values will be pulled from after program execution.
    push_config : PushConfig
        The configuration of the PushInterpreter to use when executing the program.

    """

    arity = field(type=int, mandatory=True)
    output_stacks = field(type=list, mandatory=True)
    push_config = field(type=PushConfig, mandatory=True, initial=PushConfig())


class Program(Saveable, PRecord):
    """A Push program containing all information needed to run the code with consistent behavior.

    Attributes
    ----------
    code : CodeBlock
        The Push code expressing the logic of the program.
    signature : ProgramSignature
        The specification of the program and the configuration of the Push interpreter that it was evolved under.

    """

    code = field(type=CodeBlock, mandatory=True)
    signature = field(type=ProgramSignature, mandatory=True)



"""The :mod:`atoms` module defines all types of Push Atoms.

An `Atom` is a piece of code in the Push language. Some `Atoms` only exist in `Programs` and some only exist in
`Genomes`.

A `Literal` atom is a constant value of one of the supported PushType.

An `Instruction` atom is a wrapped function that can be used to manipulate a `PushState`. These atoms are where the
logic of our Push programs is stored.

An `Input` atom is a reference to one of the Push program's inputs (aka arguments). Inputs are referenced by position.

A `CodeBlock` is a sequence of other Atoms is used to express nested block of code. They cannot appear in `Genomes`.

"""
from __future__ import annotations
from typing import Sequence, Tuple, Optional
from itertools import chain, count

from pyrsistent import PClass, field, CheckedPVector

from pyshgp.push.types import PushType


class Atom:
    """Base class of all Atoms. The fundamental element of Push programs."""

    ...


class Closer(Atom, PClass):
    """An Atom dedicated to denoting the close of a CodeBlock in a `Genomes` representation."""

    ...


class Literal(Atom, PClass):
    """An Atom which holds a constant value.

    Attributes
    ----------
    value : Any
        The value to be stored in Literal.
    push_type : PushType, optional
        The `PushType` of the `Literal`. Usually, the PushType's underlying native
        type is the same as the type of the Literal's value. The PushType is
        used to determine how to route the Literal onto the correct PushStack
        of the PushState during program execution.

    """

    __invariant__ = lambda l: (l.push_type.is_instance(l.value), "Value is not of PushType.")
    value = field(mandatory=True)
    push_type = field(type=PushType, mandatory=True)


class InstructionMeta(Atom, PClass):
    """An identifier of a Push Instruction.

    The `InstructionMeta` is a placeholder atom. When a PushInterpreter evaluates
    an instruction ID, it searches for an instruction with the given
    name in its `InstructionSet`. If found, this instruction is evaluated by the `PushInterpreter`.

    Attributes
    ----------
    name: str
        The name of the Instruction being referenced.
    code_blocks: int
        The number of code blocks that are expected to follow the instruction.

    """

    name = field(type=str, mandatory=True)
    code_blocks = field(type=int, mandatory=True)


class Input(Atom, PClass):
    """A reference to a positional argument of the `PushProgram` which the atom is part of.

    Attributes
    ----------
    input_index: int
        The position of the associated input to Push program.

    """

    input_index = field(type=int, mandatory=True)


class CodeBlock(Atom, CheckedPVector):
    """An `Atom` which holds a sequence of other atoms in order to express a nested code block."""

    __type__ = Atom

    def size(self, depth: int = 1) -> int:
        """Return the size of the block and the size of all the nested blocks."""
        return sum([el.size(depth + 1) + 1 if isinstance(el, CodeBlock) else 1 for el in self])

    def depth(self) -> int:
        """Return the farthest depth of the CodeBlock."""
        i = iter(self)
        level = 0
        try:
            for level in count():
                i = chain([next(i)], i)
                i = chain.from_iterable(s for s in i if isinstance(s, Sequence) and not isinstance(s, str))
        except StopIteration:
            return level
        return 0

    def code_at_point(self, ndx: int) -> Optional[Atom]:
        """Return a nested element of the CodeBlock using depth first traversal."""
        if ndx == 0:
            return self
        i = ndx
        for el in self:
            i = i - 1
            if i == 0:
                return el
            if isinstance(el, CodeBlock):
                next_depth = el.code_at_point(i)
                if next_depth is not None:
                    return next_depth
                i = i - el.size()
        return None

    def with_code_inserted_at_point(self, code: Atom, index: int) -> CodeBlock:
        """Insert an element into the CodeBlock using depth first traversal."""
        if index > self.size():
            return self.append(code)
        _, new = self._attempt_code_insert(code, index)
        return new

    def _attempt_code_insert(self, code: Atom, index: int) -> Tuple[bool, CodeBlock]:
        i = index
        for idx, el in enumerate(self):
            if i == 0:
                new = CodeBlock(self[:idx] + [code] + self[idx:])
                return True, new
            if isinstance(el, CodeBlock):
                inserted, new = el._attempt_code_insert(code, i - 1)
                if inserted:
                    return True, self.set(idx, new)
                i = i - (el.size() + 1)
            i = i - 1
        if i == 0:
            return True, self.append(code)
        return False, self



"""Push program execution requires configuration to ensure all Push programs are safe.

Synthesizing Push requires generating random code. When executed, this random code can contain a number of
bad behaviors such as: exponential memory utilization, infinite loops, and impractically long runtime. This
modules defines important configuration objects that are used to constrain push program execution.

A block of Push code can produce different behavior under different ``PushConfig`` settings. This implies that
the ``PushConfig`` used to produce a solution program must be a part of the definition of a push ``Program``.

"""
from collections import Sequence
from typing import Union

from pyrsistent import PRecord, field


class PushConfig(PRecord):
    """A configuration for a Push program.

    Attributes
    ----------
    step_limit : int
        Max number of atoms to process before terminating the execution of a
        given program. Default is 500
    runtime_limit : int
        Max number of seconds to run a push program before forcing
        termination. Default is 10.
    growth_cap : int
        Max number of elements that can be added to a PushState at any given
        step of program execution. If exceeded, program terminates. Default is
        500.
    collection_size_cap : int, optional
        Max size of any collection (code blocks, vectors, strings, etc). Default is 1000.
    numeric_magnitude_limit : float, optional
        Max magnitude of numbers. Default is 1 million.

    """

    step_limit = field(type=int, initial=500, mandatory=True)
    runtime_limit = field(type=int, initial=10, mandatory=True)
    growth_cap = field(type=int, initial=500, mandatory=True)
    collection_size_cap = field(type=int, initial=1000, mandatory=True)
    numeric_magnitude_limit = field(type=float, initial=1e12, mandatory=True)


def constrain_collection(config: PushConfig, coll: Sequence) -> Sequence:
    """Constrains the collection to a size that is safe for Push program execution."""
    if len(coll) > config.collection_size_cap:
        return coll[:config.collection_size_cap]
    return coll


def constrain_number(config: PushConfig, n: Union[int, float]) -> Union[int, float]:
    """Constrains the number to a magnitude that is safe for Push program execution."""
    if abs(n) > config.numeric_magnitude_limit:
        sign = -1 if n < 0 else 1
        return config.numeric_magnitude_limit * sign
    return n



"""The :mod:`interpreter` module defines the ``PushInterpreter`` used to run Push programs."""
from typing import Union
import time
from enum import Enum

from pyshgp.push.instruction import Instruction
from pyshgp.push.program import Program
from pyshgp.push.state import PushState
from pyshgp.push.instruction_set import InstructionSet
from pyshgp.push.atoms import Atom, Closer, Literal, InstructionMeta, CodeBlock, Input
from pyshgp.push.config import PushConfig
from pyshgp.tap import tap
from pyshgp.validation import PushError


class PushInterpreterStatus(Enum):
    """Enum class of all potential statuses of a PushInterpreter."""

    normal = 1
    step_limit_exceeded = 2
    runtime_limit_exceeded = 3
    growth_cap_exceeded = 4


class PushInterpreter:
    """An interpreter capable of running Push programs.

    Parameters
    ----------
    instruction_set : Union[InstructionSet, str], optional
        The ``InstructionSet`` to use for executing programs. Default is "core"
        which instantiates an ``InstructionSet`` using all the core instructions.

    Attributes
    ----------
    instruction_set : InstructionSet
        The ``InstructionSet`` to use for executing programs.
    state : PushState
        The current ``PushState``. Contains one stack for each ``PushType``
        mentioned by the instructions in the instruction set.
    status : PushInterpreterStatus
        A string denoting if the interpreter has encountered a situation
        where non-standard termination was required.

    """

    def __init__(self,
                 instruction_set: Union[InstructionSet, str] = "core",
                 reset_on_run: bool = True):
        self.reset_on_run = reset_on_run
        # If no instruction set given, create one and register all instructions.
        if instruction_set == "core":
            self.instruction_set = InstructionSet(register_core=True)
        else:
            self.instruction_set = instruction_set

        self.type_library = self.instruction_set.type_library

        # Initialize the PushState and status
        self.state: PushState = None
        self.status: PushInterpreterStatus = None
        self._validate()

    def _validate(self):
        library_type_names = set(self.type_library.keys())
        required_stacks = self.instruction_set.required_stacks() - {"stdout", "exec", "untyped"}
        if not required_stacks <= library_type_names:
            raise ValueError(
                "PushInterpreter instruction_set and type_library are incompatible. {iset} vs {tlib}. Diff: {d}".format(
                    iset=required_stacks,
                    tlib=library_type_names,
                    d=required_stacks - library_type_names,
                ))

    def _evaluate_instruction(self, instruction: Instruction, config: PushConfig):
        self.state = instruction.evaluate(self.state, config)

    def untyped_to_typed(self):
        """Infer ``PushType`` of items on state's untyped queue and push to corresponding stacks."""
        while len(self.state.untyped) > 0:
            el = self.state.untyped.popleft()
            push_type = self.type_library.push_type_of(el, error_on_not_found=True)
            self.state[push_type.name].push(el)

    @tap
    def evaluate_atom(self, atom: Atom, config: PushConfig):
        """Evaluate an ``Atom``.

        Parameters
        ----------
        atom : Atom
            The Atom (``Literal``, ``InstructionMeta``, ``Input``, or ``CodeBlock``) to
            evaluate against the current ``PushState``.
        config : PushConfig
            The configuration of the Push program being run.

        """
        try:
            if isinstance(atom, InstructionMeta):
                self._evaluate_instruction(self.instruction_set[atom.name], config)
            elif isinstance(atom, Input):
                input_value = self.state.inputs[atom.input_index]
                self.state.untyped.append(input_value)
            elif isinstance(atom, CodeBlock):
                for a in atom[::-1]:
                    self.state["exec"].push(a)
            elif isinstance(atom, Literal):
                self.state[atom.push_type.name].push(atom.value)
            elif isinstance(atom, Closer):
                raise PushError("Closers should not be in push programs. Only genomes.")
            else:
                raise PushError("Cannot evaluate {t}, require a subclass of Atom".format(t=type(atom)))
            self.untyped_to_typed()
        except Exception as e:
            err_type = type(e).__name__
            err_msg = str(e)
            raise PushError(
                "{t} raised while evaluating {atom}. Original message: \"{m}\"".format(
                    t=err_type,
                    atom=atom,
                    m=err_msg
                ))

    @tap
    def run(self,
            program: Program,
            inputs: list,
            print_trace: bool = False) -> list:
        """Run a Push ``Program`` given some inputs and desired output ``PushTypes``.

        The general flow of this method is:
            1. Create a new push state
            2. Load the program and inputs.
            3. If the exec stack is empty, return the outputs.
            4. Else, pop the exec stack and process the atom.
            5. Return to step 3.

        Parameters
        ----------
        program : Program
            Program to run.
        inputs : list
            A sequence of values to use as inputs to the push program.
        print_trace : bool
            If True, each step of program execution will be summarized in stdout.

        Returns
        -------
        Sequence
            A sequence of values pulled from the final push state. May contain
            pyshgp.utils.Token.no_stack_item if output stacks are empty.

        """
        push_config = program.signature.push_config

        if self.reset_on_run or self.state is None:
            self.state = PushState(self.type_library, push_config)
            self.status = PushInterpreterStatus.normal

        # Setup
        self.state.load_code(program.code)
        self.state.load_inputs(inputs)
        stop_time = time.time() + push_config.runtime_limit
        steps = 0

        if print_trace:
            print("Initial State:")
            self.state.pretty_print()

        # Iterate atom evaluation until entire program is evaluated.
        while len(self.state["exec"]) > 0:
            # Stopping conditions
            if steps > push_config.step_limit:
                self.status = PushInterpreterStatus.step_limit_exceeded
                break
            if time.time() > stop_time:
                self.status = PushInterpreterStatus.runtime_limit_exceeded
                break

            # Next atom in the program to evaluate.
            next_atom = self.state["exec"].pop()

            if print_trace:
                start = time.time()
                print("\nCurrent Atom: " + str(next_atom))

            # Evaluate atom.
            old_size = self.state.size()
            self.evaluate_atom(next_atom, push_config)
            if self.state.size() > old_size + push_config.growth_cap:
                self.status = PushInterpreterStatus.growth_cap_exceeded
                break

            if print_trace:
                duration = time.time() - start
                print("Current State (step {step}):".format(step=steps))
                self.state.pretty_print()
                print("Step duration:", duration)
            steps += 1

        if print_trace:
            print("Finished program evaluation.")

        return self.state.observe_stacks(program.signature.output_stacks)


DEFAULT_INTERPRETER = PushInterpreter()
